"""Framework errors."""


class Error(Exception):
    """Base error type."""


class ConfigError(Error):
    """Raised on configuration error."""


class DocumentError(Error):
    """Base class for all document related errors."""
    message = "document error"

    def __init__(self, message=None, document=None, field=None, value=None):
        """Format a document or field error message."""
        message = message or self.message
        if document:
            if isinstance(document, type):
                document = document.__name__
            elif isinstance(document, object):
                document = document.__class__.__name__
            if field:
                message = "{}: {}.{} = {}".format(message, document, field, repr(value))
            else:  # pragma: no cover
                message = "{}: {}".format(message, document)
        elif field and value:
            message = "{}: {} = {}".format(message, field, repr(value))
        elif field:  # pragma: no cover
            message = "{}: {}".format(message, field)
        elif value:
            message = "{}: {}".format(message, repr(value))
        super(DocumentError, self).__init__(message)


class OperationError(DocumentError):
    """Raised when an operation fails for a document."""
    message = "operation failed"


class ValidationError(DocumentError):
    """Raised when document or field validation fails."""
    message = "value is invalid"


class EncodingError(DocumentError):
    """Raised on field encoding/decoding error."""
    encode_message = "failed to encode value"
    decode_message = "failed to decode value"

    def __init__(self, message=None, document=None, field=None, value=None, encode=True):
        """Format an encoding error message."""
        self.message = encode and self.encode_message or self.decode_message
        super(EncodingError, self).__init__(message, document, field, value)



"""A few useful utilities."""
import re


def to_snake_case(name):
    """Convert a camel case string to snake case."""
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def get_projection(fields):
    if fields:
        return list(fields)
    return None



"""Additional data encoders."""
from .errors import EncodingError
from collections import OrderedDict


class BaseEncoder(object):
    """Base encoder class."""

    def __init__(self, document):
        """Create an encoder for the given document."""
        self.document = document

    def is_array_value(self, value):
        """Return True if the value can be encoded to an array."""
        return isinstance(value, (list, tuple, set))

    def is_array_field(self, field):
        """Return True if a field stores an array."""
        typ = getattr(field, 'typ', None)
        return typ and hasattr(typ, 'encode_element')

    def encode_default(self, name, value):
        """Return a value with default encoding."""
        return value

    def encode_builtin(self, builtin, name, value):
        try:
            return builtin(value)
        except (TypeError, ValueError):
            msg = "unable to encode {} value".format(builtin.__name__)
            raise EncodingError(msg, self.document, name, value)

    def encode_int(self, name, value):
        """Return a value encoded as an integer."""
        return self.encode_builtin(int, name, value)

    def encode_float(self, name, value):
        """Return a value encoded as a float."""
        return self.encode_builtin(float, name, value)

    def encode_bool(self, name, value):
        """Return a value encoded as a boolean."""
        return self.encode_builtin(bool, name, value)

    def encode_str(self, name, value):
        """Return a value encoded as a string."""
        return self.encode_builtin(str, name, value)

    def encode_geojson_coords(self, name, value):
        """Return a value encoded as GeoJSON coordinates."""
        if not self.is_array_value(value):
            return self.encode_float(name, value)

        encoded = []
        for item in value:
            encoded.append(self.encode_geojson_coords(name, item))
        return encoded

    def encode_geojson(self, name, value):
        """Return a value encoded as a GeoJSON value."""
        try:
            value = OrderedDict(value)
        except (TypeError, ValueError):
            msg = "unable to encode value as GeoJSON"
            raise EncodingError(msg, self.document, name, value)

        encoded = OrderedDict()
        for item_name, item_value in value.iteritems():
            if item_name == 'type':
                item_value = self.encode_str(item_name, item_value)
            elif item_name == 'coordinates':
                item_value = self.encode_geojson_coords(item_name, item_value)
            else:
                item_value = self.encode_default(item_name, item_value)
            encoded[item_name] = item_value
        return encoded


class OperatorEncoder(BaseEncoder):
    """Base encoder for specs with operators."""
    ops = {}

    def get_encode_method(self, op):
        """Return the encode method for an operator."""
        name = self.ops.get(op)
        if name:
            name = 'encode_' + name
            if hasattr(self, name):
                return getattr(self, name)
        return self.encode_default

    def get_field(self, name):
        """Return the field for the given name."""
        return self.document._meta.get_field(name)


class SortEncoder(BaseEncoder):
    """Encode sort specs."""

    def encode(self, value):
        """
        Encode a sort value. Value must be convertible to an int or OrderedDict or raises an
        EncodingError.
        """
        if value is None:
            return None

        try:
            value = OrderedDict(value)
            encoded = OrderedDict()
            for field, direction in value.iteritems():
                direction = self.encode_int(field, direction)
                field = self.encode_str(field, field)
                encoded[field] = direction
            return encoded.items()
        except (TypeError, ValueError):
            pass

        try:
            return int(value)
        except (TypeError, ValueError):
            pass

        raise EncodingError('unable to encode sort value', value=value)


class QueryEncoder(OperatorEncoder):
    """Encode query specs."""

    ops = {
        '$gt': 'field',
        '$gte': 'field',
        '$lt': 'field',
        '$lte': 'field',
        '$ne': 'field',
        '$in': 'field',
        '$nin': 'field',
        '$or': 'logical',
        '$and': 'logical',
        '$not': 'negation',
        '$nor': 'logical',
        '$exists': 'bool',
        '$type': 'int',
        '$mod': 'mod',
        '$regex': 'str',
        '$options': 'str',
        '$search': 'str',
        '$language': 'str',
        '$where': 'str',
        '$all': 'field',
        '$elemMatch': 'array_query',
        '$size': 'int',
        '$geoWithin': 'geo',
        '$geoIntersects': 'geo',
        '$near': 'geo',
        '$nearSphere': 'geo',
    }

    def __init__(self, document):
        """Create a query encoder for a document type."""
        self.document = document

    def is_operator_name(self, name):
        """Return True if the name is an operator name."""
        return str(name)[:1] == '$'

    def is_operator_value(self, value):
        """Return True if a value contains operators."""
        try:
            value = OrderedDict(value)
            for name in value.keys():
                if self.is_operator_name(name):
                    return True
        except (TypeError, ValueError):
            pass
        return False

    def is_compiled_regex(self, value):
        """Return True if a value is a compiled regular expression."""
        return hasattr(value, 'pattern')

    def encode_logical(self, name, value):
        """Return a value encoded for a logical operator."""
        if not self.is_array_value(value):
            raise EncodingError("unable to encode logical operator", field=name, value=value)
        return [self.encode(v) for v in value]

    def encode_negation(self, name, value):
        """Return a value encoded for negation."""
        return self.encode(value)

    def encode_mod(self, name, value):
        """Return a value encoded for modulus division."""
        if not self.is_array_value(value):
            raise EncodingError("unable to encode mod operator", field=name, value=value)
        return [self.encode_float(name, v) for v in value]

    def encode_array_query(self, name, value):
        """Return a value encoded as an array."""
        from .types import DocumentType
        field = self.get_field(name)
        if field and self.is_array_field(field) and isinstance(field.typ.typ, DocumentType):
            document = field.typ.typ.document
        else:
            from .document import Document
            document = Document
        return QueryEncoder(document).encode(value)

    def encode_geo(self, name, value):
        """Return a value encoded as a geo query."""
        try:
            value = OrderedDict(value)
        except (TypeError, ValueError):
            raise EncodingError("unable to encode geo query", self.document, name, value)

        encoded = OrderedDict()
        for item_name, item_value in value.iteritems():
            if item_name == '$geometry':
                item_value = self.encode_geojson(item_name, item_value)
            elif item_name == '$maxDistance':
                item_value = self.encode_int(item_name, item_value)
            else:
                item_value = self.encode_default(item_name, item_value)
            encoded[item_name] = item_value
        return encoded

    def encode_field(self, name, value):
        """Return a value encoded as a field value."""
        if value is None:
            return None
        if self.is_compiled_regex(value):
            return self.encode_default(name, value)

        field = self.get_field(name)
        if field:
            if self.is_array_value(value):
                if self.is_array_field(field):
                    value = field.encode(self.document, name, value)
                else:
                    value = [field.encode(self.document, name, v) for v in value]
            else:
                if self.is_array_field(field):
                    value = field.typ.encode_element(self.document, name, value)
                else:
                    value = field.encode(self.document, name, value)
        else:
            value = self.encode_default(name, value)
        return value

    def encode_operator(self, name, value):
        """Return a value encoded as an operator dictionary."""
        if value is None:
            return None
        encode_method = self.get_encode_method(name)
        return encode_method(name, value)

    def encode_operators(self, name, value):
        """Return a value encoded as an operator dictionary."""
        encoded = OrderedDict()
        for item_name, item_value in value.iteritems():
            item_name = self.encode_str(name, item_name)
            if item_value is not None:
                encode_method = self.get_encode_method(item_name)
                item_value = encode_method(name, item_value)
            encoded[item_name] = item_value
        return encoded

    def encode(self, value):
        """Return an encoded query value."""
        if not value:
            return None
        try:
            value = OrderedDict(value)
        except (TypeError, ValueError):
            raise EncodingError("unable to encode query", self.document, '<query>', value)

        encoded = OrderedDict()
        for item_name, item_value in value.iteritems():
            item_name = self.encode_str('<query>', item_name)
            if self.is_operator_name(item_name):
                item_value = self.encode_operator(item_name, item_value)
            if self.is_operator_value(item_value):
                item_value = self.encode_operators(item_name, item_value)
            else:
                item_value = self.encode_field(item_name, item_value)
            encoded[item_name] = item_value
        return encoded


class UpdateEncoder(OperatorEncoder):
    """Encode update specs."""

    ops = {
        '$inc': 'scalar',
        '$mul': 'scalar',
        '$rename': 'str',
        '$setOnInsert': 'scalar',
        '$set': 'scalar',
        '$unset': 'unset',
        '$min': 'scalar',
        '$max': 'scalar',
        '$currentDate': 'currentdate',
        '$addToSet': 'add',
        '$pop': 'int',
        '$pullAll': 'array',
        '$pull': 'query',
        '$pushAll': 'array',
        '$push': 'push',
        '$bit': 'bitwise',
    }

    def __init__(self, document):
        """Create an encoder for the given document class."""
        self.document = document

    def is_positional(self, name):
        """Return True if the field name has a positional operator attached."""
        return name[-2:] == '.$'

    def get_field_name(self, name):
        """Return a clean field name."""
        if self.is_positional(name):
            return name[:-2]
        return name

    def get_field(self, name):
        """Return a named document field."""
        name = self.get_field_name(name)
        return super(UpdateEncoder, self).get_field(name)

    def encode_scalar(self, name, value):
        """Encode a scalar update value."""
        if self.is_positional(name):
            return self.encode_array_element(name, value)
        field = self.get_field(name)
        if field:
            return field.encode(self.document, name, value)
        return self.encode_default(name, value)

    def encode_unset(self, name, value):
        """Encode an unset update value."""
        return ''

    def encode_array(self, name, values):
        """Encode an array update value."""
        field = self.get_field(name)
        if field:
            if self.is_array_value(values):
                if self.is_array_field(field):
                    values = field.encode(self.document, name, values)
                else:
                    values = [field.encode(self.document, name, value) for value in values]
            else:
                if self.is_array_field(field):
                    values = field.typ.encode_element(self.document, name, values)
                else:
                    values = field.encode(self.document, name, values)
        else:
            values = self.encode_default(name, values)
        return values

    def encode_array_element(self, name, value):
        """Encode an array element update value."""
        field = self.get_field(name)
        if field:
            if self.is_array_field(field):
                return field.typ.encode_element(self.document, name, value)
            else:
                return field.typ.encode(self.document, name, value)
        return self.encode_default(name, value)

    def encode_add(self, name, value):
        """Encode an value for adding to a set."""
        if isinstance(value, dict):
            encoded = OrderedDict()
            for item_name, item_value in value.iteritems():
                if item_name == '$each':
                    encoded[item_name] = self.encode_array(name, item_value)
                else:
                    encoded[item_name] = self.encode_default(name, item_value)
            value = encoded
        else:
            value = self.encode_array_element(name, value)
        return value

    def encode_sort(self, name, value):
        """Encode a sort spec."""
        return SortEncoder(self.document).encode(value)

    def encode_push(self, name, value):
        """Encode a push update value."""
        if isinstance(value, dict):
            encoded = OrderedDict()
            for item_name, item_value in value.iteritems():
                if item_name == '$each':
                    encoded[item_name] = self.encode_array(name, item_value)
                elif item_name == '$slice':
                    encoded[item_name] = self.encode_int(name, item_value)
                elif item_name == '$sort':
                    encoded[item_name] = self.encode_sort(name, item_value)
                elif item_name == '$position':
                    encoded[item_name] = self.encode_int(name, item_value)
                else:
                    encoded[item_name] = self.encode_default(name, item_value)
            value = encoded
        else:
            value = self.encode_array_element(name, value)
        return value

    def encode_query(self, name, value):
        """Encode a query update value."""
        from .query import Query
        if not isinstance(value, Query):
            value = Query(value)
        return value.encode(self.document)

    def encode_currentdate(self, name, value):
        """Encode a currentdate value."""
        if isinstance(value, dict):
            encoded = OrderedDict()
            for item_name, item_value in value.iteritems():
                if item_name == '$type':
                    encoded[item_name] = self.encode_str(name, item_value)
                else:
                    encoded[item_name] = self.encode_default(name, item_value)
            value = encoded
        else:
            value = bool(value)
        return value

    def encode_bitwise(self, name, value):
        """Encode a bitwise value."""
        try:
            value = OrderedDict(value)
        except (TypeError, ValueError):
            raise EncodingError("unable to encode bitwise value", self.document, name, value)

        encoded = OrderedDict()
        for item_name, item_value in value.iteritems():
            encoded[str(item_name)] = self.encode_int(name, item_value)
        return encoded

    def encode(self, value):
        """Return an encoded update value."""
        if not value:
            return None
        try:
            value = OrderedDict(value)
        except (TypeError, ValueError):
            raise EncodingError("unable to encode update", self.document, '<update>', value=value)

        encoded = OrderedDict()
        for op, values in value.iteritems():
            encode_method = self.get_encode_method(op)
            encoded_update = OrderedDict()
            for name, value in values.iteritems():
                encoded_update[name] = encode_method(name, value)
            encoded[op] = encoded_update
        return encoded



"""Query tools."""
from .encoders import QueryEncoder
from collections import OrderedDict
from copy import deepcopy


class Query(object):
    """A query abstracts a MongoDB query."""

    def __init__(self, criteria):
        """
        Initialize the query with the provided criteria. The criteria may be a Query or any value
        that may be converted to an OrderedDict which is the internal query representation. Raises
        a TypeError if criteria does not fit these requirements.
        """
        if criteria is None:
            criteria = OrderedDict()
        elif isinstance(criteria, Query):
            criteria = criteria.criteria.copy()
        elif not isinstance(criteria, OrderedDict):
            try:
                criteria = OrderedDict(criteria)
            except (TypeError, ValueError):
                raise TypeError("invalid query value {}".format(repr(criteria)))
        self.criteria = criteria

    def copy(self):
        """Return a copy of the query."""
        return Query(deepcopy(self.criteria))

    def encode(self, document, raw=None):
        """Return the encoded query in the context of the given document."""
        if raw:
            return self.criteria
        return QueryEncoder(document).encode(self.criteria)

    def _op(self, op, query):
        """Combine two queries with an operator and return the resulting query."""
        if len(self.criteria) == 0:
            return query.copy()
        if len(query.criteria) == 0:
            return self.copy()
        if len(self.criteria) == 1 and op in self.criteria:
            criteria = deepcopy(self.criteria)
            criteria[op].append(deepcopy(query.criteria))
            return Query(criteria)
        else:
            criteria = [(op, [deepcopy(self.criteria), deepcopy(query.criteria)])]
            return Query(criteria)

    def negate(self):
        """Negate a query."""
        if len(self.criteria) == 0:
            return self.copy()
        if len(self.criteria) == 1 and '$not' in self.criteria:
            return Query(deepcopy(self.criteria['$not']))
        else:
            return Query({'$not': deepcopy(self.criteria)})

    def nor(self, query):
        """Nor this query with another and return the resulting query."""
        return self._op('$nor', query)

    def __and__(self, query):
        """Logically and this query (left side) with the provided query (right side)."""
        return self._op('$and', query)

    def __or__(self, query):
        """Logically or this query (left side) with the provided query (right side)."""
        return self._op('$or', query)

    def __eq__(self, query):
        """Return True if the query is equal to this one."""
        if isinstance(query, Query):
            return self.criteria == query.criteria
        return False


Q = Query



from .connection import Connection, configure
from .document import Document
from .errors import (
    ConfigError,
    DocumentError,
    EncodingError,
    Error,
    OperationError,
    ValidationError,
)
from .field import Field
from .query import Q, Query
from .reference import Reference
from bson import ObjectId

__all__ = [
    'ConfigError',
    'Connection',
    'Document',
    'DocumentError',
    'EncodingError',
    'Error',
    'Field',
    'ObjectId',
    'OperationError',
    'Q',
    'Query',
    'Reference',
    'ValidationError',
    'configure',
]



"""Define field types that may be encoded and/or decoded."""
from collections import OrderedDict
from datetime import date, datetime, time
from errors import EncodingError

epoch = date(1970, 1, 1)
registered_field_types = []


def is_type(cls, typ):
    if not isinstance(cls, type):
        return False
    return issubclass(cls, typ)


def is_field_type(typ):
    """Return True if the type looks like a FieldType."""
    return isinstance(typ, FieldType)


def is_date_type(typ):
    """Return True if type is a date."""
    return is_type(typ, date) and not is_type(typ, datetime)


def is_datetime_type(typ):
    """Return True if type is a datetime."""
    return is_type(typ, datetime)


def is_time_type(typ):
    """Return True if type is a time."""
    return is_type(typ, time)


def is_document_type(typ):
    """Return True if type is a Document."""
    from .document import Document
    return is_type(typ, Document)


def is_list_type(typ):
    """Return True if obj is a list or tuple."""
    return is_type(typ, list) or is_type(typ, tuple)


def is_set_type(typ):
    """Return True if obj is a set."""
    return is_type(typ, set)


def is_dict_type(typ):
    """Return True if obj is a dict."""
    return is_type(typ, dict)


def is_date_obj(obj):
    """Return True if obj is a date."""
    return isinstance(obj, date) and not isinstance(obj, datetime)


def is_datetime_obj(obj):
    """Return True if obj is a datetime."""
    return isinstance(obj, datetime)


def is_time_obj(obj):
    """Return True if obj is a time."""
    return isinstance(obj, time)


def is_document_obj(obj):
    """Return True if obj is a Document."""
    from .document import Document
    return isinstance(obj, Document)


def is_list_obj(obj):
    """Return True if obj is a list or tuple."""
    return isinstance(obj, (list, tuple))


def is_set_obj(obj):
    """Return True if obj is a set."""
    return isinstance(obj, set)


def is_dict_obj(obj):
    """Return True if obj is a dict."""
    return isinstance(obj, dict)


def register_field_type(check, field_type):
    """
    Register a field type. The check is called on typ and should return True if the field type
    supports the type. The field_type is the field type class.
    """
    registered_field_types.append((check, field_type))


class FieldType(object):
    """Abstract class used for defining a field type."""

    @classmethod
    def create(cls, typ):
        if is_field_type(typ):
            return typ
        for check, field_type in registered_field_types:
            if check(typ):
                return field_type(typ)
        return BuiltinType(typ)

    def __init__(self, *args, **kwargs):
        pass

    def encode(self, cls, name, value):
        """Return the value encoded for storage in the database."""
        return value

    def decode(self, cls, name, value):
        """Return the value decoded from database storage."""
        return value

    def validate(self, cls, name, value):
        """Raise ValidationError if the field fails to validate."""


class BuiltinType(FieldType):
    """Used for built-in Python types which can be called with a single argument."""

    def __init__(self, builtin):
        """Create a field type using the given builtin type."""
        if isinstance(builtin, type) and issubclass(builtin, basestring):
            builtin = unicode
        self.builtin = builtin

    def encode(self, cls, name, value):
        """Return the encoded value."""
        try:
            return self.builtin(value)
        except (TypeError, ValueError):
            msg = "failed to encode value as {}".format(self.builtin.__class__.__name__)
            raise EncodingError(msg, cls, name, value, True)


class DateType(FieldType):
    """Support date values."""

    def encode(self, cls, name, value):
        """Return a date value as a datetime."""
        if is_datetime_obj(value):
            value = value.date()
        if is_date_obj(value):
            value = datetime.combine(value, time(0))
        if not is_datetime_obj(value):
            raise EncodingError(None, cls, name, value, True)
        return value

    def decode(self, cls, name, value):
        """Return the date value from the stored datetime."""
        if not is_datetime_obj(value):
            raise EncodingError(None, cls, name, value, False)
        return value.date()


class DateTimeType(FieldType):
    """Support datetime values."""

    def encode(self, cls, name, value):
        """Convert time and date values into a datetime."""
        if is_date_obj(value):
            value = datetime.combine(value, time(0))
        elif is_time_obj(value):
            value = datetime.combine(epoch, value)
        elif not is_datetime_obj(value):
            raise EncodingError(None, cls, name, value, True)
        return value

    def decode(self, cls, name, value):
        """Return the datetime value for the stored value."""
        if not is_datetime_obj(value):
            raise EncodingError(None, cls, name, value, False)
        return value


class TimeType(FieldType):
    """Support time values."""

    def encode(self, cls, name, value):
        """Convert a datetime into a time."""
        if is_datetime_obj(value):
            value = value.time()
        if is_time_obj(value):
            value = datetime.combine(epoch, value)
        else:
            raise EncodingError(None, cls, name, value, True)
        return value

    def decode(self, cls, name, value):
        """Return the time value from the stored datetime."""
        if is_datetime_obj(value):
            return value.time()
        raise EncodingError(None, cls, name, value, False)


class DocumentType(FieldType):
    """Support subdocuments."""

    def __init__(self, document):
        """Create a document type object with the given document class."""
        self.document = document

    def encode(self, cls, name, value):
        """Return the value encoded as a raw subdocument."""
        if is_document_obj(value):
            return value._encode()
        raise EncodingError(None, cls, name, value, True)

    def decode(self, cls, name, value):
        """Return the value decoded as a subdocument object."""
        if hasattr(value, 'get'):
            return self.document._decode(value)
        raise EncodingError(None, cls, name, value, False)

    def validate(self, cls, name, value):
        """Raise ValidationError if the field fails to validate."""
        self.document._validate(value)


class ListType(FieldType):
    """Support a list of typed values."""

    def __init__(self, typ):
        """Create a list type using the given type."""
        if is_list_type(typ) or is_list_obj(typ) and len(typ) == 0:
            self.typ = None
        elif is_list_obj(typ):
            self.typ = FieldType.create(typ[0])

    def encode_element(self, cls, name, value):
        """Return the encoded value for a single list element."""
        if self.typ:
            return self.typ.encode(cls, name, value)
        return value

    def encode(self, cls, name, value):
        """Return the value encoded as a list of encoded values."""
        if self.typ is not None:
            encoded = []
            for item in value:
                encoded.append(self.encode_element(cls, name, item))
            return encoded
        return list(value)

    def decode(self, cls, name, value):
        """Return the value decoded as a list of decoded values."""
        if self.typ is not None:
            decoded = []
            for item in value:
                decoded.append(self.typ.decode(cls, name, item))
            return decoded
        return list(value)


class SetType(FieldType):
    """Support a set of typed values."""

    def __init__(self, typ):
        """Create a set type using the given type."""
        if is_set_type(typ) or is_set_obj(typ) and len(typ) == 0:
            self.typ = None
        elif is_set_obj(typ):
            self.typ = FieldType.create(list(typ)[0])

    def encode_element(self, cls, name, value):
        """Return the encoded value for a single set element."""
        if self.typ:
            return self.typ.encode(cls, name, value)
        return value

    def encode(self, cls, name, value):
        """Return the value encoded as a set of encoded values."""
        if self.typ is not None:
            encoded = []
            for item in value:
                encoded.append(self.encode_element(cls, name, item))
            return encoded
        return list(value)

    def decode(self, cls, name, value):
        """Return the value decoded as a list of decoded values."""
        if self.typ is not None:
            decoded = set()
            for item in value:
                decoded.add(self.typ.decode(cls, name, item))
            return decoded
        return set(value)


class DictType(FieldType):
    """Support a list of type dict values with strings for keys."""

    def __init__(self, typ):
        """Create s dict type using the given type."""
        if is_dict_type(typ) or is_dict_obj(typ) and len(typ) == 0:
            self.typ = None
        elif is_dict_obj(typ):
            self.typ = FieldType.create(typ.values()[0])

    def encode(self, cls, name, value):
        """Return the value encoded as a dict of encoded values."""
        if self.typ is not None:
            encoded = OrderedDict()
            for key, item in value.iteritems():
                encoded[str(key)] = self.typ.encode(cls, name, item)
            return encoded
        return OrderedDict(value)

    def decode(self, cls, name, value):
        """Return the value decoded as a list of decoded values."""
        if self.typ is not None:
            decoded = OrderedDict()
            for key, item in value.iteritems():
                decoded[key] = self.typ.decode(cls, name, item)
            return decoded
        return OrderedDict(value)


register_field_type(is_date_type, DateType)
register_field_type(is_datetime_type, DateTimeType)
register_field_type(is_time_type, TimeType)
register_field_type(is_document_type, DocumentType)
register_field_type(lambda t: is_list_type(t) or is_list_obj(t), ListType)
register_field_type(lambda t: is_set_type(t) or is_set_obj(t), SetType)
register_field_type(is_dict_type, DictType)



"""Configure active databases."""
from .errors import ConfigError
from pymongo import MongoClient
from pymongo.errors import AutoReconnect
from time import sleep

connections = {}


class Connection(object):
    """A lazy database connection."""

    @classmethod
    def configure(cls, config):
        """
        Create a new connection using the provided configuration. The configuration may be a uri or
        a dictionary containing values for uri, prefix. retries, and backoff. Additional options
        are passed to the MongoClient init.
        """
        if not isinstance(config, dict):
            config = {'uri': config}
        uri = config.pop('uri', None)
        if not uri:
            raise ConfigError("invalid uri: {}".format(uri))
        return cls(uri, **config)

    def __init__(self, uri, prefix=None, retries=None, backoff=None, **options):
        """
        Initialize the connection. The URI should be a full mongodb:// URI including the database
        name. The prefix is optional and is a string used to prefix all collection names. The
        retries are the number of times to retry a command when pymongo raises an AutoReconnect.
        The backoff is how much time to add to the time between each successive retry. Additional
        args are passed to pymongo's MongoClient().
        """
        if retries is None or retries < 0:
            retries = 4
        if backoff is None or backoff < 0:
            backoff = 0.5
        self.uri = uri
        self.prefix = prefix or ''
        self.retries = retries
        self.backoff = backoff
        self.options = options
        self._client = None
        self._database = None

    @property
    def client(self):
        """Return the client used by this connection."""
        if self._client is None:
            self._client = MongoClient(self.uri, **self.options)
        return self._client

    @property
    def database(self):
        """Return the database used by this connection."""
        if self._database is None:
            self._database = self.client.get_default_database()
        return self._database

    def __getitem__(self, name):
        """Return a collection from the connection's database."""
        if not name:
            return None
        return CollectionProxy(self, self.database[self.prefix + name])

    def autoreconnect(self, func):
        """Return the provided function decorated with autoreconnect functionality."""
        def reconnect(*args, **kwargs):
            tries = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except AutoReconnect:
                    if tries >= self.retries:
                        raise
                sleep(tries * self.backoff)
                tries += 1
        return reconnect

    def close(self):
        """Close the connection."""
        if self._client:
            self._client.close()


class CollectionProxy(object):
    """Proxy method calls to collections in order to handle AutoReconnect errors."""

    def __init__(self, connection, collection):
        """Create a proxy around the provided collection."""
        self._connection = connection
        self._collection = collection

    def __getattr__(self, name):
        """Return an attribute. Methods are decorated with autoreconnect functionality."""
        value = getattr(self._collection, name)
        if hasattr(value, '__call__'):
            value = self._connection.autoreconnect(value)
        return value


def get(name):
    """Return a named connection or None if not found."""
    return connections.get(name)


def add(name, connection=None, default=False):
    """
    Add a named connection. The connection may be a Connection object or configuration for one.
    If a connection string/object is not specified, `mongodb://localhost//{{name}}` is used.

    When default is True, the global default connection is updated. This is used
    by all models which lack an explicitly-defined connection.
    """
    if connection is None:
        connection = Connection.configure('mongodb://localhost/{}'.format(name))
    if not isinstance(connection, Connection):
        connection = Connection.configure(connection)
    if default:
        connections[None] = connection
    connections[name] = connection


def configure(config):
    """
    Initialize connections from the provided configuration. The config object is a dictionary
    containing named connections. The keys of the dictionary are the names while the values are the
    configuration for each connection. The value may be a string containing the database URI or a
    dictionary with the keys 'uri' for the database URI, and 'prefix' for a string with which to
    prefix collection names. Remaining options are passed as keyword args to pymongo's
    MongoClient.
    """
    for name, options in config.iteritems():
        add(name, options)



"""Cursors for iterating over documents."""
from .query import Query
from .utils import get_projection


class Cursor(object):
    """
    Cursors are used to iterate over multiple documents or to further refine the results of a
    find().
    """

    def __init__(self, document, collection, query, fields, raw, **options):
        """
        Initialize the cursor with the given find query. The find will be executed against the
        given connection. Additional args are passed to pymongo's find().
        """
        self.document = document
        self.collection = collection
        self.query = self._make_query(query)
        self.fields = fields
        self.raw = raw
        self.options = options
        self.options.pop('manipulate', None)

    def _make_query(self, query):
        """Return a query for an object of dubious origin."""
        if query is not None and not isinstance(query, Query):
            query = Query(query)
        return query

    @property
    def _criteria(self):
        """Return the encoded criteria for the cursor."""
        if self.query is None:
            return None
        return self.query.encode(self.document, self.raw)

    @property
    def connection(self):
        """Return the connection for the cursor."""
        return self.collection._connection

    @property
    def pymongo(self):
        """Return the pymongo cursor which underlies this object."""
        if not getattr(self, '_pymongo_cursor', None):
            self._pymongo_cursor = self.collection.find(
                self._criteria, projection=get_projection(self.fields), **self.options)
        return self._pymongo_cursor

    def count(self):
        """Count the number of objects matching this cursor."""
        return self.pymongo.count()

    def find(self, query):
        """Refine the cursor's scope with an additional query. Return a new cursor."""
        query = self.query & self._make_query(query)
        return Cursor(self.document, self.collection, query, self.fields, self.raw, **self.options)

    def remove(self):
        """Remove the documents matched by this cursor."""
        res = self.collection.remove(self._criteria)
        return res.get('n', 0)

    def __getitem__(self, index):
        """Return the document at the given index."""
        return self.document._decode(self.pymongo[index], get_projection(self.fields))

    def __iter__(self):
        """Return the cursor iterator."""
        return self

    def __len__(self):
        return self.pymongo.count()

    def next(self):
        """Return the next item in the iterator."""
        def decode_next():
            return self.document._decode(self.pymongo.next(), self.fields)
        return self.connection.autoreconnect(decode_next)()

    def close(self):
        """Explicitly close the cursor."""
        if getattr(self, '_pymongo_cursor', None):
            self._pymongo_cursor.close()
            self._pymongo_cursor = None



"""Field objects."""
from .errors import EncodingError
from .types import FieldType


class BaseField(object):
    """Base field object which all fields inherit from."""
    default = None

    def getter(self, obj, name):
        """Return a document attribute as this field."""
        if name not in obj._attrs:
            value = obj._raw.get(name)
            if value is not None:
                try:
                    value = self.decode(obj.__class__, name, value)
                except EncodingError as e:
                    raise TypeError(e)
            obj._attrs[name] = value
        return obj._attrs[name]

    def setter(self, obj, name, value):
        """Set a document attribute as this field."""
        obj._attrs[name] = value
        obj._dirty.add(name)
        if obj._partial:
            obj._partial.add(name)

    def __call__(field, doc, name):
        """Return the document property used to access the field."""
        @property
        def prop(self):
            return field.getter(self, name)

        @prop.setter
        def prop(self, value):
            return field.setter(self, name, value)

        return prop

    def encode(self, cls, name, value):
        """Return the value encoded for storage in the database."""
        return value

    def decode(self, cls, name, value):
        """Return the value decoded from database storage."""
        return value

    def validate(self, cls, name, value):
        """Validate the field value. Raise ValidationError on failure."""


class Field(BaseField):
    """A field object defines how a document field behaves."""

    def __init__(self, typ, require=False, default=None, strict=True, modifier=None):
        """
        Initialize a new field. typ is the type of the field and may be a FieldType or a built-in
        Python type. require is a boolean that indicates whether or not the field is required.
        default is the default value to be stored if the field's value is None. strict determines
        whether or not type conversion and validation should occur during storage and retrieval of
        the field. modifier is a callable used to modify the value before insert or update. It
        takes the current field value as its only parameter. The return value is encoded and saved.
        """
        self.typ = FieldType.create(typ)
        self.require = require
        self.default = default
        self.strict = strict
        self.validators = [self.typ.validate]
        self.modifier = modifier

    def ensure(self, func):
        """
        Ensure the field value passes the provided validation function. The validation function
        takes three arguments: the document, the field name, and the field value. It raises
        ValidationError if validation failes.
        """
        self.validators.append(func)

    def encode(self, cls, name, value):
        """Return the value encoded for storage in the database."""
        if self.strict:
            value = self.typ.encode(cls, name, value)
        return value

    def decode(self, cls, name, value):
        """Return the value decoded from database storage."""
        if self.strict:
            value = self.typ.decode(cls, name, value)
        return value

    def validate(self, cls, name, value):
        """Validate the field value. Raise ValidationError on failure."""
        if self.strict:
            for validator in self.validators:
                validator(cls, name, value)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

try:
    readme = open(os.path.join(here, 'README.md')).read()
except IOError:
    readme = ''

install_requires = [
    'pymongo>=2.7.0',
]
setup_requires = [
    'coverage>=3.7.0',
    'nose>=1.3.0',
]

setup(
    name='bearfield',
    version='1.9.1',
    description="Small MongoDB object layer.",
    long_description=readme,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: BSD License",
        "Topic :: Database",
    ],
    keywords='python mongodb',
    author='WiFast',
    author_email='rgb@wifast.com',
    url='https://github.com/WiFast/bearfield',
    license='BSD-derived',
    zip_safe=False,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=setup_requires,
    test_suite = 'nose.collector',
    entry_points='',
)



"""Document and subdocument classes."""
from .cursor import Cursor
from .encoders import SortEncoder, UpdateEncoder
from .errors import OperationError, ValidationError
from .meta import DocumentBuilder
from .query import Query
from .utils import get_projection


class Document(object):
    """
    A document or subdocument. Document properties are defined in an optional Meta subclass. In
    order for a document to be saved to a database it must associate itself with a named connection
    by setting the 'connection' meta attribute to that connection name. The collection name is the
    connection prefix plus the snake cased class name of the document. This may be overridden by
    setting the 'collection' property to the desired name of the collection. The connection's
    prefix will still be prepended to the name.

    Fields are defined on the document by assigning Field objects to class attributes. See the
    Field class for details on field parameters.

    A document may be provided as the type to a Field. This will cause that field to be treated as
    a subdocument.
    """
    __metaclass__ = DocumentBuilder

    @classmethod
    def _decode(cls, raw, fields=None):
        """
        Return a document decoded from a MongoDB record. If fields is not None then a partial
        document will be created with those field values.
        """
        if raw is None:
            return None
        doc = cls()
        doc._raw = raw.copy()
        doc._partial = cls._meta.get_partial(fields)
        return doc

    @classmethod
    def _validate(cls, raw, partial=None, update=False):
        """
        Validate the raw document. Raise a ValidationError if validation fails. If the raw document
        is an update document then update should be set to True.
        """
        if update:
            raw = raw.get('$set', {})
        required = []
        for name, field in cls._meta.get_fields(partial).iteritems():
            value = raw.get(name)
            if value is None:
                if field.require:
                    required.append(name)
            else:
                field.validate(cls, name, value)

        if not update and required:
            doc = cls.__name__
            required = ', '.join(sorted(required))
            raise ValidationError("{} is missing required fields: {}".format(doc, required))

    @classmethod
    def find(cls, query=None, fields=None, connection=None, raw=None, sort=None, **options):
        """
        Query the database for documents. Return a cursor for further refining or iterating over
        the results. If fields is not None only return the field values in that list. Additional
        args are passed to pymongo's find().
        """
        collection = cls._meta.get_collection(connection)
        fields = cls._meta.get_partial(fields)
        if not raw:
            sort = SortEncoder(cls).encode(sort)
        return Cursor(cls, collection, query, fields, raw, sort=sort, **options)

    @classmethod
    def find_one(cls, query=None, fields=None, connection=None, raw=None, sort=None, **options):
        """
        Query the database for a single document. Return the document or None if not found.
        Additional args are passed to pymongo's find(). If fields is not None only return the field
        values in that list.
        """
        collection = cls._meta.get_collection(connection)
        fields = cls._meta.get_partial(fields)
        options.pop('manipulate', None)
        criteria = Query(query).encode(cls, raw)
        if not raw:
            sort = SortEncoder(cls).encode(sort)
        return cls._decode(collection.find_one(criteria, projection=get_projection(fields),
                                               sort=sort, **options), fields)

    @classmethod
    def find_and_modify(cls, query, update, fields=None, connection=None, raw=None, sort=None,
                        new=None, **options):
        """
        Query the database for a document, update it, then return the old document before
        modification. Additional args are passed to pymongo's find_and_modify().
        """
        if cls._meta.disable_update:
            msg = "updates to {} are disabled".format(cls.__class__.__name__)
            raise OperationError(msg)

        collection = cls._meta.get_collection(connection)
        fields = cls._meta.get_partial(fields)
        criteria = Query(query).encode(cls, raw)
        if new is None:
            new = False
        if not raw:
            sort = SortEncoder(cls).encode(sort)
            update = UpdateEncoder(cls).encode(update)
        raw = collection.find_and_modify(criteria, update, projection=get_projection(fields),
                                         new=new, sort=sort, **options)
        return cls._decode(raw, fields)

    @classmethod
    def count(cls, connection=None):
        """Count the number of objects in this collection."""
        collection = cls._meta.get_collection(connection)
        return collection.count()

    def __init__(self, *args, **kwargs):
        """Initialize the document with values."""
        for name, value in kwargs.iteritems():
            setattr(self, name, value)

    def _encode(self, update=False):
        """
        Return the document as a dictionary suitable for saving. If update is
        True then an update document is returned.
        """
        def modify(name, field):
            if getattr(field, 'modifier', None):
                setattr(self, name, field.modifier(getattr(self, name)))

        raw = {}
        if update:
            sets = {}
            unsets = {}
            for name, field in self._meta.get_fields(self._partial).iteritems():
                modify(name, field)
                if name not in self._dirty:
                    continue
                value = self._attrs.get(name)
                if value is None:
                    unsets[name] = ""
                else:
                    sets[name] = field.encode(self.__class__, name, value)
            if sets:
                raw['$set'] = sets
            if unsets:
                raw['$unset'] = unsets
        else:
            for name, field in self._meta.get_fields(self._partial).iteritems():
                modify(name, field)
                if name in self._attrs:
                    value = self._attrs[name]
                    if value is not None:
                        value = field.encode(self.__class__, name, value)
                else:
                    value = self._raw.get(name)
                if value is not None:
                    raw[name] = value
        return raw

    def _reset(self, raw):
        """Reset internal field storage using the raw document."""
        self._raw.update(raw)
        self._attrs = {}
        self._dirty = set()

    def save(self, connection=None, **options):
        """
        Save the model to the database. Effectively performs an insert if the _id field is None and
        a full document update otherwise. Additional args are passed to pymongo's save().
        """
        if self._meta.disable_save:
            msg = "saves to {} are disabled".format(self.__class__.__name__)
            raise OperationError(msg)

        if self._partial:
            raise OperationError("unable to save partial document")

        collection = self._meta.get_collection(connection)
        raw = self._encode()
        self._validate(raw, self._partial)
        options.pop('manipulate', None)
        self._id = collection.save(raw, manipulate=True, **options)
        self._reset(raw)

    def insert(self, connection=None, **options):
        """
        Insert the document. This ignores the state of the _id field and forces an insert. This may
        necessitate setting _id to None prior to calling insert. Though this could be used to
        insert the same document into multiple databases. Additional args are passed to pymongo's
        insert().
        """
        if self._meta.disable_insert:
            msg = "inserts to {} are disabled".format(self.__class__.__name__)
            raise OperationError(msg)

        collection = self._meta.get_collection(connection)
        raw = self._encode()
        self._validate(raw, self._partial)
        options.pop('manipulate', None)
        self._id = collection.insert(raw, manipulate=True, **options)
        self._reset(raw)

    def update(self, update=None, connection=None, raw=None, sort=None, **options):
        """
        Update the document in the database using the provided update statement. If update is None
        (the default) an update statement is created to set all of the dirty fields in the
        document. This uses the _id field to find the document to update and will raise an error if
        no _id is set. Additional args are passed to pymongo's update(). Return True if an update
        was performed or False if no update was needed.
        """
        if self._meta.disable_update:
            msg = "updates to {} are disabled".format(self.__class__.__name__)
            raise OperationError(msg)

        if not self._id:
            raise OperationError("unable to update document without an _id")

        collection = self._meta.get_collection(connection)

        if not raw:
            sort = SortEncoder(self.__class__).encode(sort)

        if not update:
            update = self._encode(True)
        elif not raw:
            update = UpdateEncoder(self.__class__).encode(update)

        self._validate(update.get('$set', {}), self._partial, True)
        if update:
            options.pop('multi', None)
            options.pop('new', None)
            options.pop('fields', None)
            res = collection.find_and_modify(
                {'_id': self._id}, update, projection=get_projection(self._partial), multi=False,
                new=True, sort=sort, **options)
            self._reset(res)
            return True
        return False

    def remove(self, connection=None, **options):
        """
        Remove the document from the database. Additional args are passed to pymongo's remove().
        Return True if the document was removed or False if there was nothing to remove.
        """
        if self._meta.disable_remove:
            msg = "removal of {} is disabled".format(self.__class__.__name__)
            raise OperationError(msg)

        collection = self._meta.get_collection(connection)
        if self._id:
            res = collection.remove(self._id)
            return res.get('n', 0) > 0
        return False



"""Code for managing references."""
from .document import Document
from .errors import EncodingError
from .field import BaseField
from .query import Query
from bson import ObjectId
from bson.errors import InvalidId


class Reference(BaseField):
    """
    A reference is effectively a stored query against a known collections. A reference may store
    one of two values: an ObjectId or a Query object. A reference may be set using those two values
    or something else. If the value is a string an attempt is made to convert it to an ObjectId.
    An attempt is made to convert any other type to a Query. A TypeError is raised if conversion
    fails.
    """

    def __init__(self, doctype, require=True, default=None):
        """
        Initialize a new reference. The reference uses a Document class to determine which
        collection to make queries against.
        """
        self.doctype = doctype
        self.require = require
        self.default = default

    def getter(self, obj, name):
        """Return a ReferenceFinder for the reference."""
        return ReferenceFinder(self, obj, name)

    def setter(self, obj, name, value):
        """
        Set the Reference query. Raise TypeError on value conversion failure. Raise ValueError if
        the value is a Document with no _id field.
        """
        from .document import Document
        if value is not None:
            if isinstance(value, basestring):
                try:
                    value = ObjectId(value)
                except InvalidId:
                    raise TypeError("'{}' is not a valid ObjectId".format(value))
            elif isinstance(value, Document):
                _id = getattr(value, '_id', None)
                if not _id:
                    raise ValueError(
                        "Document of type {} has no _id".format(value.__class__.__name__))
                value = _id
            elif not isinstance(value, (Query, ObjectId)):
                value = Query(value)
        super(Reference, self).setter(obj, name, value)

    def encode(self, cls, name, value):
        """Return an encoded ObjectID or Query."""
        if isinstance(value, Query):
            value = value.encode(self.doctype)
        elif isinstance(value, Document):
            value = value._id
        return value

    def decode(self, cls, name, value):
        """Return a decoded ObjectID or Query."""
        if not isinstance(value, ObjectId):
            try:
                value = Query(value)
            except TypeError:
                raise EncodingError("invalid Query or ObjectId: {}".format(repr(value)))
        return value


class ReferenceFinder(object):
    """Proxy calls to the reference value."""

    def __init__(self, reference, document, name):
        self.reference = reference
        self.document = document
        self.name = name

    @property
    def value(self):
        """Return the reference value."""
        if self.name not in self.document._attrs:
            value = self.document._raw.get(self.name)
            if value is not None:
                try:
                    value = self.reference.decode(self.document.__class__, self.name, value)
                except EncodingError as e:
                    raise TypeError(e)
            self.document._attrs[self.name] = value
        return self.document._attrs[self.name]

    @property
    def query(self):
        """
        Return the reference as a query. If the reference is None then it returns the query
        {'_id': None}.
        """
        value = self.value
        if value is None or isinstance(value, ObjectId):
            value = Query({'_id': value})
        return value

    def find(self, fields=None, connection=None, **options):
        """
        Return the results of a find using the reference. Options are the same as Document.find()
        without query.
        """
        cls = self.reference.doctype
        return cls.find(self.query, fields, connection, **options)

    def find_one(self, fields=None, connection=None, **options):
        """
        Return the results of a find_one using the reference. Options are the same as
        Document.find_one() without query.
        """
        cls = self.reference.doctype
        return cls.find_one(self.query, fields, connection, **options)



"""Meta functionality used for document creation."""
from .connection import Connection, get as get_connection
from .errors import OperationError
from .field import BaseField, Field
from bson import ObjectId
from utils import to_snake_case


class DocumentMeta(object):
    """Metadata container for Document classes."""

    def __init__(self, cls, attrs, meta):
        """Initialize new document class metadata."""
        self.cls = cls
        self.fields = {}
        self.options = {}

        self.bind_init()

        fields = {}
        for base in reversed(cls.__bases__):
            if isinstance(getattr(base, '_meta', None), self.__class__):
                fields.update(base._meta.fields)

        if attrs:
            for name, attr in attrs.items():
                if isinstance(attr, BaseField):
                    fields[name] = attr

        self.fields = fields

        if meta:
            self.options.update(vars(meta))

        self.connection = self.options.pop('connection', None)
        self.collection = self.options.pop('collection', None)
        if not self.collection:
            self.collection = to_snake_case(cls.__name__)

        self.subdocument = not bool(self.connection)
        if not self.subdocument and '_id' not in self.fields:
            self.fields['_id'] = Field(ObjectId, require=False)

        self.bind_fields()

    def bind_init(meta):
        """Bind init hook to the document class."""
        parent = meta.cls.__init__
        # prevent recursive decoration
        if hasattr(parent, 'parent'):
            parent = parent.parent

        def __init__(self, *args, **kwargs):
            self._raw = {}
            for name, default in meta.defaults.iteritems():
                if hasattr(default, '__call__'):
                    field = self._meta.get_field(name)
                    default = field.encode(meta.cls, name, default())
                self._raw[name] = default
            self._attrs = {}
            self._dirty = set()
            self._partial = None
            return parent(self, *args, **kwargs)

        __init__.name = parent.__name__
        __init__.hook = True
        __init__.parent = parent
        meta.cls.__init__ = __init__

    def bind_fields(self):
        """Bind fields to the document class."""
        defaults = {}
        for name, field in self.fields.items():
            setattr(self.cls, name, field(self.cls, name))
            default = field.default
            if default is not None:
                if not hasattr(default, '__call__'):
                    default = field.encode(self.cls, name, field.default)
                field.validate(self.cls, name, default)
            defaults[name] = default

        self.defaults = defaults

    def get_connection(self, connection=None):
        """
        Return the connection associated with this document. If connection is provided then it will
        be used instead. This value may be the name of a connection or an actual connection object.

        If this document has no connection (`self.connection is None`),
        the default global connection is used.
        """
        if connection:
            if isinstance(connection, Connection):
                return connection
            else:
                return get_connection(str(connection))
        return get_connection(self.connection)

    def get_collection(self, connection=None):
        """
        Return the collection associated with this document. If error it True then errors are Raise
        an OperationError if the document has no collection.
        """
        if self.collection:
            connection = self.get_connection(connection)
            if connection:
                return connection[self.collection]
        raise OperationError(
            "document {} has no connection, and no default exists".format(self.cls.__name__))

    def get_partial(self, fields):
        """Return a valid partial value from a list of fields."""
        if fields:
            return {'_id'} | set(fields)
        return None

    def get_fields(self, partial):
        """Return a dictionary containing active fields."""
        if partial:
            return {k: self.fields[k] for k in partial if k in self.fields}
        return self.fields

    def get_field(self, name):
        """Return the named field. Supports dot syntax to retrieve fields from subdocuments."""
        from .types import DocumentType
        names = name.split('.')
        names.reverse()
        doc = self.cls
        field = None

        while len(names):
            name = names.pop()
            field = doc._meta.fields.get(name)
            if not hasattr(field, 'typ') or not isinstance(field.typ, DocumentType):
                break
            doc = field.typ.document

        if len(names):
            return None
        return field

    @property
    def readonly(self):
        """Return True if the document is readonly."""
        return bool(self.options.get('readonly', False) or (
            self.options.get('disable_save', False) and
            self.options.get('disable_insert', False) and
            self.options.get('disable_update', False) and
            self.options.get('disable_remove', False)))

    @property
    def disable_save(self):
        """Return True if save is disabled for the document."""
        return (bool(self.options.get('readonly', False) or
                self.options.get('disable_save', False)))

    @property
    def disable_insert(self):
        """Return True if insert is disabled for the document."""
        return (bool(self.options.get('readonly', False) or
                self.options.get('disable_insert', False)))

    @property
    def disable_update(self):
        """Return True if update is disabled for the document."""
        return (bool(self.options.get('readonly', False) or
                self.options.get('disable_update', False)))

    @property
    def disable_remove(self):
        """Return True if remove is disabled for the document."""
        return (bool(self.options.get('readonly', False) or
                self.options.get('disable_remove', False)))


class DocumentBuilder(type):
    """Metaclass for building document classes."""

    def __new__(meta, name, bases, attrs):
        """Create and attach metadata to the document."""
        Meta = attrs.pop('Meta', {})
        cls = type.__new__(meta, name, bases, attrs)
        cls._meta = DocumentMeta(cls, attrs, Meta)
        return cls
